#!/opt/homebrew/bin/python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"

########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2021.09.14                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION: Created to sepparate the crazy logic between repeating and single occurance threads. The flow of this #
#    class is Sleep -> Action, where as the repeating class is (Action -> Sleep) <- REPEAT.                            #
#   BUGS:                                                                                                              #
#   FUTURE:                                                                                                            #
#                                                                                                                      #
########################################################################################################################


from . import BaseThread


class DelayThread(BaseThread):
	def __init__(self, name: str, *, action: callable, time: callable):
		BaseThread.__init__(self, name, action=action, time=time)


	def __call__(self) -> None:
		try:  # make it safe!!!
			self.sleep(self.time())
			if(self._is_active):
				self.action()

		except Exception as error:
			print("ERROR")
			print(error)




import mpzinke
import types
import typing
from typing import Any, Dict, Optional


class Validator:
	def __init__(self, argument_types: Dict[str, Any], descendent_class: Optional[type]=None):
		"""
		
		"""
		# Must be `type(self)` because an inheritted object returns true for `isinstance(self, Validator)`
		assert(type(self) is not Validator), "'Validator' must be inheritted"
		assert(issubclass(type(self), Validator)), f"'{type(self).__name__}' must be of type 'Validator'"

		descendent_class = descendent_class or Validator.class_with_Validator_parent(type(self))
		parameter_types = descendent_class.__init__.__annotations__
		Validator.check_for_missing_arguments(argument_types, parameter_types)
		Validator.check_argument_types(argument_types, parameter_types)


	@staticmethod
	def check_for_missing_arguments(argument_types: dict, parameter_types: dict) -> None:
		"""
		SUMMARY: Check if any parameter is not supplied in the arguments values.
		RAISES: KeyError
		"""
		missing_arguments_strings: list[str] = [] 
		for name, type in parameter_types.items():
			if(name not in argument_types):
				missing_arguments_strings.append(f"'{name}' of type '{type}'")

		if(len(missing_arguments_strings) != 0):
			raise KeyError(f"""Missing key(s) for {", ".join(missing_arguments_strings)}""")


	@staticmethod
	def check_argument_types(argument_types: dict, parameter_types: dict) -> None:
		"""
		Check if any parameter is not of proper type.
		RAISES: ValueError.
		"""
		failed_param_strings: list[str] = []
		for name, required_type in parameter_types.items():
			if(not Validator.check_argument_type(argument_types[name], required_type)):
				failed_argument_type = type(argument_types.get(name)).__name__
				failed_string = f"'{name}' must be of type '{required_type.__name__}' not '{failed_argument_type}'"
				failed_param_strings.append(failed_string)

		if(len(failed_param_strings) != 0):
			raise ValueError(", ".join(failed_param_strings))


	@staticmethod
	def check_argument_type(value: Any, needed_type: type) -> bool:
		"""
		https://stackoverflow.com/questions/49171189/whats-the-correct-way-to-check-if-an-object-is-a-typing-generic
		"""
		if((origin := typing.get_origin(needed_type)) is types.UnionType or origin is typing.Union):
			return Validator.union_types_match(value, needed_type)

		elif(isinstance(needed_type, (mpzinke.Generic, types.GenericAlias, typing._GenericAlias))):
			return Validator.generic_types_match(value, needed_type)

		# int
		return isinstance(value, needed_type)


	@staticmethod
	def generic_types_match(argument: Any, needed_type: type) -> bool:
		"""
		EG. `list[<__args__[0]>] or Dict[<__args__[0]>, <__args__[1]]`
		"""
		if(not isinstance(argument, typing.get_origin(needed_type))):  # if(not isinstance({1: ['1', '1'], ...}, dict))
			return False

		# If the number of types for the generic is one, iterate through those types
		if(len((generics_args := needed_type.__args__)) == 1):
			return all(Validator.check_argument_type(subargument, generics_args[0]) for subargument in argument)

		# {1: ['1', '1'], 2: ['2', '2'], 3: ['3', '3']}
		for argument_children in (argument.items() if(isinstance(argument, dict)) else argument):
			# (1, int), (['1', '1'], list[str])
			for grandchild, type in zip(argument_children, generics_args):
				if(not Validator.check_argument_type(grandchild, type)):
					return False

		return True


	@staticmethod
	def union_types_match(value: Any, needed_type: type) -> bool:
		"""
		EG. `int|str``
		"""
		return any(Validator.check_argument_type(value, unioned_type) for unioned_type in needed_type.__args__)


	@staticmethod
	def class_with_Validator_parent(class_type: type) -> type:
		for base in class_type.__bases__:
			if(base is not None):
				if(base == Validator):
					return class_type

				if((ancestor := Validator.class_with_Validator_parent(base)) is not None):
					return ancestor

		return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"

########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2022.04.14                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION:                                                                                                       #
#   BUGS:                                                                                                              #
#   FUTURE:                                                                                                            #
#                                                                                                                      #
########################################################################################################################


from datetime import datetime, timedelta
import requests
from typing import Optional
import warnings
from warnings import warn as Warn


class DownloadIterator:
	"""
	SUMMARY: Used to wrap the response's iter_content iterator.
	DETAILS: Has iter and next functions to initialize and support iteration, wrapping the iter_content method.
	NOTES:   Based on:
	           https://towardsdatascience.com/how-to-loop-through-your-own-objects-in-python-1609c81e11ff
	           https://gist.github.com/zmwangx/0dcbdbe6f67f3540dd73e777ef1b2a89
	"""

	def __init__(self, response, *, timeout: Optional[int]=None, callback: Optional[callable]=None,
		chunk_size: Optional[int]=1024, suppress_warning: bool=False
	):
		"""
		SUMMARY: Contructs the DownloadIterator object.
		PARAMS:  Takes the HTTP reponse object, the timeout amount, OPTIONALLY: a callback function to run each
		         iteration of the loop, the chunk_size to collect from the response.
		"""
		self._callback: callable = callback  # Function pointer called in case you want to add something
		self._downloaded: int = 0  # The amount of byte download
		self._end: Optional[datetime] = None  # The timestamp of the max allotted time to download data (from iteration start)
		self._chunk_size: int = chunk_size  # The disired chunk size to download at once
		self._response: requests.Response = response  # The response for which the content download iterator is called
		self._response_content_iterator: iter = None  # The iter_content object
		self._suppress_warning: bool = suppress_warning
		self._timeout: Optional[int] = timeout  # The stored timeout amount

		warnings.formatwarning = DownloadIterator.warning_message


	def __iter__(self):
		"""
		SUMMARY: Initializes the iteration process for the DownloadIterator.
		DETAILS: Sets the timeout timestamp from the moment iteration initialization is called. Initialize iteration of
		         response's content.
		RETURNS: It's as the object being iterated over.
		"""
		if(self._timeout):
			self._end = datetime.now() + timedelta(seconds=self._timeout)
		elif(not self._suppress_warning):
			Warn(f"No time out specified. This can cause the download to run indefinitely.")

		self._response_content_iterator = iter(self._response.iter_content(chunk_size=self._chunk_size))
		return self


	def __next__(self) -> bytes:
		"""
		SUMMARY: Gets the next value for the iteration & checks whether the download elapsed time has timed out.
		DETAILS: Checks whether the download elapsed time has timed out. Calls optional callback function to do any
		         additional processes. Increments the downloaded amount.
		RETURNS: The next data downloaded.
		THROWS:  Timeout Exception if too much time has elapsed.
		"""
		if(self._timeout and self._end <= datetime.now()):
			error_message = f"Download for response has exceeded time amount of {self._timeout}." \
			  + f" Downloaded {self._downloaded} bytes"
			raise TimeoutError(error_message)

		if(self._callback is not None):  # Allow other optional actions
			self._callback(self)

		self._downloaded += self._chunk_size
		return next(self._response_content_iterator)


	@staticmethod
	def warning_message(message, *_: list) -> str:
		"""
		SUMMARY: Used to override warning format & return only the warning message.
		"""
		return str(message)+"\n"



#!/opt/homebrew/bin/python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"

########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2023.07.31                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION:                                                                                                       #
#   BUGS:                                                                                                              #
#   FUTURE:                                                                                                            #
#                                                                                                                      #
########################################################################################################################


from flask import Flask, jsonify
from flask_cors import CORS
import sys
import traceback
from typing import Dict, Optional, TypeVar
from werkzeug.exceptions import Forbidden, HTTPException, Unauthorized


from .Route import Route, ArgMapping, HTTPMethod, URL


Server = TypeVar("Server")


class Server:
	"""
	Flask API Server.
	Methods:
		- __init__:
	"""
	EXCEPTIONS = {401: Unauthorized, 403: Forbidden}


	def __init__(self, *, authorization: Optional[callable]=None, additional_args: Optional[ArgMapping]=None,
		debug: bool=False, handle_error: Optional[callable]=None, host: str="0.0.0.0", name: str="Flask App",
		port: int=8080, version: str="1.0.0", **kwargs: dict
	):
		"""
		PARAMS:
			`additional_args: Optional[ArgMapping]=None`: An optional dictionary of args to pass to HTTPCallbacks.
			`authorization: Optional[callable]=None`: An optional callback used to authorize incoming requests.
			`handle_error: Optional[callable]=None`:  An optional callback used to handle error responses on requests.
			`host: str="0.0.0.0"`: The host on which server shall run.
			`name: str="Flask App"`: The name of the app.
			`port: int=8080`: The port on which the server shall run.
			`version: str="1.0.0"`: The version of the app.
			`**kwargs: dict`: Any additional args that will be passed to the Flask server.
		"""
		self._app = Flask(name, **kwargs)
		self._app.debug = debug

		self._cors = CORS(self._app)
		self._app.config['CORS_HEADERS'] = 'Content-Type'
		self._app.register_error_handler(Exception, self._handle_error)
		self._app.after_request(self._after_request)

		self._additional_args: Optional[ArgMapping] = {Server: self, **(additional_args or {})}
		self._authorization: callable = authorization
		self._debug: bool = debug
		self._handle_error = handle_error or self._handle_error
		self._host: str = host
		self._port: int = port
		self._routes: list[Route] = []
		self._version: str = version


	# ———————————————————————————————————————————————————— THREAD ———————————————————————————————————————————————————— #

	def __call__(self, *, debug: Optional[bool]=None, host: Optional[str]=None, port: Optional[int]=None) -> None:
		"""
		SUMMARY: Adds routes to server & class, and starts the server instance.
		DETAILS: Sets routes using hardcoded routes, functions & HTTP request methods. Calls the Flask::run method.
		"""
		if(debug is None):
			debug = self._app.debug

		if(host is None):
			host = self._host

		if(port is None):
			port = self._port

		self._app.run(debug=debug, host=host, port=port)


	def __iter__(self) -> list[str]:
		yield from {
			endpoint: doc_string for route in self._routes for endpoint, doc_string in dict(route).items()
		}.items()


	# ——————————————————————————————————————————————— REQUEST HANDLING ——————————————————————————————————————————————— #

	def _after_request(self, response):
		"""
		FROM: https://stackoverflow.com/a/30717205
		"""
		response.headers["Version"] = self._version
		return response


	def debug(self, flag: bool=True) -> None:
		self._app.debug = flag


	def _handle_error(self, error):
		"""
		SUMMARY: Handles the return response for any server error that occurs during a request.
		PARAMS:  Takes the error that has occured.
		RETURNS: The JSON of the error.
		FROM: https://readthedocs.org/projects/pallet/downloads/pdf/latest/
		 AND: https://stackoverflow.com/a/29332131
		"""
		if isinstance(error, HTTPException):
			return jsonify(error=str(error)), error.code

		try:
			exception_traceback = traceback.format_exc()
		except:
			exception_traceback = "Unknown traceback"

		print(str(error), exception_traceback, file=sys.stderr)
		return jsonify(error=str(error), traceback=exception_traceback), 500


	# ————————————————————————————————————————————————————— AUTH ————————————————————————————————————————————————————— #

	@staticmethod
	def bearer_auth(*authorization_args: list) -> callable:
		"""
		SUMMARY: Function used for no authorization. Can be called statically or dynamically.
		PARAMS:  Takes either the bearer token or the object calling it and the bearer token.
		"""
		use_message = ("Use either statically with `Server.bearer_auth(auth: str)`"
			" or dynamically with `<server_object>.bearer_auth(auth: str)`")
		if(len(authorization_args) == 0):
			raise TypeError(f"An authorization string must be passed to `bearer_auth`\n{use_message}")

		if(len(authorization_args) > 2):
			raise TypeError(f"Too many arguments passed\n{use_message}")

		if(len(authorization_args) == 1):
			if(not isinstance(authorization_args[0], str)):
				raise TypeError(f"Invalid argument passed\n{use_message}")

			authorization = authorization_args[0]

		elif(len(authorization_args) == 2):
			if(not isinstance(authorization_args[0], Server) or not isinstance(authorization_args[1], str)):
				raise TypeError(f"Invalid argument passed\n{use_message}")

			authorization = authorization_args[1]

		def check_auth(request) -> bool:
			if("Authorization" not in request.headers):
				raise Unauthorized()

			elif(request.headers["Authorization"] != f"Bearer {authorization}"):
				raise Forbidden()

		return check_auth


	@staticmethod
	def no_auth(*_: list) -> True:
		"""
		Function used for no authorization. Can be called statically or dynamically.
		"""
		return True


	# ———————————————————————————————————————————————————— ROUTES ———————————————————————————————————————————————————— #
	# ———————————————————————————————————————————————————————————————————————————————————————————————————————————————— #

	def route(self, url: URL, GET: callable=None, *, additional_args: Optional[ArgMapping]=None,
	  authorization: Optional[callable]=None, **method_mappings: Dict[HTTPMethod, callable]
	) -> None:
		"""
		SUMMARY: 
		PARAMS:  
		DETAILS: 
		"""
		# Set the properly supplied args by precedence.
		for args in [additional_args, self._additional_args, {}]:
			if((additional_args := args) is not None):
				break

		method_mappings = {method.upper(): function for method, function in method_mappings.items()}

		if(GET is not None):  # Use the GET argument
			if("GET" in method_mappings):  # Ensure 'GET' is not doubly supplied
				raise Exception(f"Ambiguous supplying of argument 'GET' and keyword argument 'GET' for URL '{url}'")

			method_mappings["GET"] = GET

		authorization = authorization or self._authorization
		self += Route(url, self, additional_args=additional_args, authorization=authorization, **method_mappings)


	def __iadd__(self: Server, route: Route) -> Server:
		if(not isinstance(route, Route)):
			raise TypeError(f"Cannot add an object of type '{type(route)}' to Server")

		route._server = self
		self._routes.append(route)

		# Set URLs for both urls that do and do not end with '/', with the exception of the root URL
		# Get the url without and with the ending '/', then remove the blank urls (ie if the root url is provided)
		urls = set(filter(None, [route._url.rstrip("/"), f"{route._url:/<1}"]))
		[self._app.add_url_rule(url, url, route, methods=list(route._methods)) for url in urls]

		return self



#!/opt/homebrew/bin/python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"

########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2023.07.31                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION:                                                                                                       #
#   BUGS:                                                                                                              #
#   FUTURE:                                                                                                            #
#                                                                                                                      #
########################################################################################################################


import asyncio
from flask import request
from inspect import iscoroutinefunction
import re
from typing import Dict, Optional, TypeVar


ArgMapping = Dict[type, any]
HTTPMethod = str
HTTPCallback = callable
HTTPMethodMapping = Dict[HTTPMethod, HTTPCallback]
URL = str


Route = TypeVar("Route")
Server = TypeVar("Server")


class Route:
	def __init__(self, url: URL, server: Optional[Server]=None, *, additional_args: Optional[ArgMapping]=None,
		authorization: Optional[callable]=None, **method_mappings: HTTPMethodMapping
	):
		self._additional_args: Dict[type, any] = additional_args or {}
		self._authorization: Optional[callable] = authorization
		self._methods: HTTPMethodMapping = {method.upper(): callback for method, callback in method_mappings.items()}
		self._server: Optional[Server] = server
		self._url: URL = url

		self._validate_HTTP_methods()
		self._validate_method_callbacks()


	def __call__(self, **kwargs: dict) -> str:
		"""
		SUMMARY: 
		PARAMS:  
		DETAILS: 
		RETURNS: 
		"""
		if(self._authorization is not None):
			self._authorization(request)

		# Add additional args to method call.
		for type, value in self._additional_args.items():
			if(type in (params := self._methods[request.method].__annotations__).values()):
				kwargs[list(params.keys())[list(params.values()).index(type)]] = value

		callback = self._methods[request.method]
		if(iscoroutinefunction(callback)):
			return asyncio.run(callback(**kwargs))

		return callback(**kwargs)


	def __iter__(self) -> Dict[str, str]:
		url = self._url
		yield from {
			f"{method} {url}": (function.__doc__ or "").strip() for method, function in self._methods.items()
		}.items()


	# ————————————————————————————————————————— ROUTES::CALLBACK  VALIDATION ————————————————————————————————————————— #

	@staticmethod
	def compare_function_params(func1: Dict[str, type], func2: Dict[str, type]) -> Dict[str, type]:
		"""
		SUMMARY: Compares the args and types of two functions.
		PARAMS:  Takes the args and types of function 1 and the args and types of function 2.
		DETAILS: Iterates through the functions' parameters. Checks whether each of the parameters names and types matches.
		RETURNS: A dictionary of {<param_name>: [<function1_type>, <function2_type>]} for mismatched parameters.
		"""
		mismatched_function_params = {}
		for param in set(list(func1) + list(func2)):
			if(param != "return" and (param not in func1 or param not in func2 or func1[param] != func2[param])):
				mismatched_function_params[param] = [func[param] if(param in func) else None for func in [func1, func2]]

		return mismatched_function_params


	def params_for_url(self) -> Dict[str, type]:
		"""
		SUMMARY: Determines the parameters and types for a URL.
		PARAMS:  Takes the URL to determine the parameters for.
		DETAILS: Determines the parameters and their types with a RegEx. Orders the parameters and types into a
		         dictionary.
		RETURNS: A dictionary of {<parameter>: <type>}.
		"""
		params = re.findall(r"<(int|string):([_a-zA-Z][_a-zA-Z0-9]*)>", self._url)
		return {param: {"int": int, "string": str}[type] for type, param in params}


	def _format_exception_string_for_bad_params(self, http_method: str, unknown_params: Dict[str, list[type]]) -> str:
		"""
		SUMMARY: 
		PARAMS:  
		DETAILS: 
		RETURNS: 
		"""
		callback_name = self._methods[http_method].__name__
		callback_params = self._methods[http_method].__annotations__

		message_strings = [f"""'{http_method}' callback '{callback_name}' is in compatable with URL '{self._url}'."""]
		if(len(params_missing_from_callback := [param for param in unknown_params if(param not in callback_params)])):
			message_strings.append(f"""'{"', '".join(params_missing_from_callback)}' missing from '{callback_name}'.""")

		url_params: Dict[str, type] = self.params_for_url()
		if(len(params_missing_from_url := [param for param in unknown_params if(param not in url_params)])):
			message_strings.append(f"""Params '{"', '".join(params_missing_from_url)}' missing from '{self._url}'.""")

		mismatched_types = [param for param in unknown_params if(param in url_params and param in callback_params)]
		if(len(mismatched_types_str := "', '".join(mismatched_types))):
			string = f"""'{http_method}' callback is missing arg(s) '{mismatched_types_str}' for URL '{self._url}'."""
			message_strings.append(string)

		return " ".join(message_strings)


	def _validate_HTTP_methods(self) -> None:
		"""
		SUMMARY: 
		PARAMS:  
		DETAILS: 
		RETURNS: 
		"""
		http_methods = ["CONNECT", "DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "POST", "PUT", "TRACE"]

		# Ensure all methods are correct HTTP methods.
		bad_methods: list[str] = [method for method in self._methods if(method.upper() not in http_methods)]
		if((bad_methods_string := "', '".join(bad_methods)) != ""):
			raise Exception(f"Method(s) '{bad_methods_string}' not an HTTP method for URL '{self._url}'")

		if(len(self._methods) == 0):  # Ensure at least 1 method supplied
			raise Exception(f"At least one HTTP method must be supplied for URL '{self._url}")


	def _validate_method_callbacks(self) -> None:
		"""
		SUMMARY: 
		PARAMS:  
		DETAILS: 
		RETURNS: 
		"""
		# Ensure all set methods have a callback (as opposed to a different type being passed)
		for http_method, callback in self._methods.items():
			if(not hasattr(callback, '__call__')):
				url = self._url
				message = f"""'{http_method}' arg must be of type 'callable', not '{type(callback)}' for URL '{url}'"""
				raise Exception(message)

		callbacks = list(self._methods.values())
		if(any(iscoroutinefunction(callback) != iscoroutinefunction(callbacks[0]) for callback in callbacks)):
			raise Exception("Callbacks must be either all synchronous or all asynchronous.")

		# Ensure all set methods have a callback (as opposed to a different type being passed)
		for http_method, callback in self._methods.items():
			if(not callable(callback)):
				url = self._url
				message = f"""'{http_method}' arg must be of type 'callable', not '{type(callback)}' for URL '{url}'"""
				raise Exception(message)

		url_params = self.params_for_url()
		for http_method, callback in self._methods.items():
			additional_arg_types: list[list[type]] = list(self._additional_args.keys())

			callback_params = callback.__annotations__
			unknown_params: Dict[str, list[Optional[type]]] = Route.compare_function_params(callback_params, url_params)

			# If any url param type is not None or callback param is not in additional arg types
			if(any(types[1] is not None for types in unknown_params.values())
			or any(types[0] not in additional_arg_types for types in unknown_params.values())):
				message = self._format_exception_string_for_bad_params(http_method, unknown_params)
				raise Exception(message)



#!/opt/homebrew/bin/python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"

########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2023.04.08                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION:                                                                                                       #
#   BUGS:                                                                                                              #
#   FUTURE:                                                                                                            #
#                                                                                                                      #
########################################################################################################################


from typing import Any, Optional


class Generic:
	"""
	A class to set runtime generics for classes, methods, and functions.
	Uses:
		- Inheritance:
			```
			class A(Generic):
				...
			
			a_int = A[int]()
			a.int = 5			
			```
		- Static Method:
			```
			class A:
				@Generic
				def static(__args__, ...):  # __args__ is the type of the generic
					...

			A.static[int](...)
			```

		- Function:
			```
			@Generic
			def function(__args__):
				...

			function[int](...)
			```
	"""
	def __class_getitem__(cls, __args__):
		if(not isinstance(__args__, tuple)):
			__args__ = (__args__,)

		assert(all(isinstance(arg, type) for arg in __args__)), "All Generic arguments must be a type"

		name = f"""{cls.__name__}[{", ".join(arg.__name__ for arg in __args__)}]"""
		return type(name, (cls,), {"__args__": __args__})


	def get_or_set__args__(self, new__args__: Optional[Any]=None) -> Optional[Any]:
		__args___name = self.__args__[0].__name__
		if(new__args__ is None):
			return getattr(self, f"_{__args___name}")

		if(not isinstance(new__args__, self.__args__[0])):
			value_type_str = type(new__args__).__name__
			message = f"'__args__Option::{__args___name}' must be of type '{__args___name}' not '{value_type_str}'"
			raise Exception(message);

		setattr(self, f"_{__args___name}", new__args__)


	# ———— WRAPPER ———— #

	def __init__(self, function: callable):
		self._function = function
		self.__annotations__ = {arg: type for arg, type in function.__annotations__.items() if(arg != "__args__")}
		self.__doc__ = function.__doc__
		self.__name__ = function.__name__



	def __call__(self, *args: list, **kwargs: dict):
		return self._function(self.__args__, *args, **kwargs)


	def __getitem__(self, __args__) -> callable:
		if(not isinstance(__args__, tuple)):
			__args__ = (__args__,)

		typed_function = Generic[__args__](self._function)
		if(self.__doc__ is not None):
			typed_function.__doc__ = self.__doc__.format(__args___names=[__arg__.__name__ for __arg__ in __args__])

		return typed_function



#!/opt/homebrew/bin/python3
# -*- coding: utf-8 -*-
__author__ = "MPZinke"


########################################################################################################################
#                                                                                                                      #
#   created by: MPZinke                                                                                                #
#   on 2020.11.04                                                                                                      #
#                                                                                                                      #
#   DESCRIPTION:                                                                                                       #
#   BUGS:                                                                                                              #
#   FUTURE: Consider moving to multiprocessing::Process instead.                                                       #
#                                                                                                                      #
########################################################################################################################


from typing import Optional


from . import BaseThread


class LoopingThread(BaseThread):
	def __init__(self, name: str, *, action: callable, time: callable):
		BaseThread.__init__(self, name, action=action, time=time)


	def __call__(self):
		try:  # make it safe!!!
			while(self._is_active):
				self.action();
				self.sleep(self.time());

		except Exception as error:
			print(error)
